from PIL import Image
import pylzma
import struct
import argparse
import re
import yara
import os


def parse_arguments():
    parser = argparse.ArgumentParser("PNG file extractor")
    parser.add_argument("-f", "--file", required=False, help="PNG file path")
    parser.add_argument(
        "-d", "--directory", required=False, help="PNGs directory file path"
    )
    parser.add_argument(
        "-o", "--output-directory", required=True, help="Output directory"
    )
    return parser.parse_args()


def parse_png(file):
    img = Image.open(file)
    pixels = img.load()
    width, height = img.size
    red_array = list()

    for y in range(height):
        for x in range(width):
            r, g, b = pixels[x, y]
            red_array.append(r)

    return bytes(red_array)


def yara_scan(data):
    rules = dict()
    match_list = list()
    rules_directory = "rules/"
    for filename in os.listdir(rules_directory):
        if filename.endswith(".yar") or filename.endswith(".yara"):
            file = rules_directory + os.sep + filename
            rules[filename] = file
            extension = os.path.splitext(file)[1]
    try:
        compiledrules = yara.compile(
            filepaths=rules,
            externals={"filename": filename, "filepath": file, "extension": extension},
        )
        matches = compiledrules.match(data=data)
        for match in matches:
            yara_sig = str(match)
            match_list.append(yara_sig)
            print("\t[*] yara rule matched: {}".format(yara_sig))
    except yara.Error as e:
        print(e)


def parse_structure(file_path, output_directory):

    red_bytes = parse_png(file_path)
    (comp_size,) = struct.unpack("<I", red_bytes[0:4])
    uncompressed_data = pylzma.decompress(red_bytes[4:comp_size], maxlength=comp_size)

    matches = re.compile(b"MZ\x90\x00").finditer(uncompressed_data)
    for m in matches:
        output_filename = ""
        payload_offset = m.start()
        (payload_size,) = struct.unpack(
            "<I", uncompressed_data[payload_offset - 4:payload_offset]
        )
        payload = uncompressed_data[payload_offset:payload_offset + payload_size]
        output_filename = os.path.join(
            output_directory, os.path.basename(file_path) + "_" + hex(payload_offset)
        )
        print(
            "[+] Payload at offset: 0x{:2X},  Payload size: 0x{:2X} saved as: {}".format(
                payload_offset, payload_size, output_filename
            )
        )
        yara_scan(payload)

        with open(output_filename, "wb") as f:
            f.write(payload)


def for_each_file_in_directory(path, output_directory):
    if not os.path.isdir(path):
        raise RuntimeError("Path is not a directory")

    for (dirpath, _, filenames) in os.walk(path):
        for file in filenames:
            try:
                file_path = os.path.join(dirpath, file)
                parse_structure(file_path, output_directory)

            except Exception as e:
                print("[-] Exception: {}".format(e))
                continue


def main():
    print(
        """
______                   _  _                                      _                    _                _                      _               
| ___ \                 | || |                                    | |                  | |              | |                    | |              
| |_/ /__ _  _ __  __ _ | || |  __ _ __  __   _ __    __ _  _   _ | |  ___    __ _   __| |    ___ __  __| |_  _ __  __ _   ___ | |_  ___   _ __ 
|  __// _` || '__|/ _` || || | / _` |\ \/ /  | '_ \  / _` || | | || | / _ \  / _` | / _` |   / _ \\ \/ /| __|| '__|/ _` | / __|| __|/ _ \ | '__|
| |  | (_| || |  | (_| || || || (_| | >  <   | |_) || (_| || |_| || || (_) || (_| || (_| |  |  __/ >  < | |_ | |  | (_| || (__ | |_| (_) || |   
\_|   \__,_||_|   \__,_||_||_| \__,_|/_/\_\  | .__/  \__,_| \__, ||_| \___/  \__,_| \__,_|   \___|/_/\_\ \__||_|   \__,_| \___| \__|\___/ |_|   
                                             | |             __/ |                                                                             
                                             |_|            |___/                                                                              
        """
    )
    args = parse_arguments()
    if not os.path.exists(args.output_directory):
        os.makedirs(args.output_directory)
        
    if args.file:
        parse_structure(args.file, args.output_directory)
    elif args.directory:
        for_each_file_in_directory(args.directory, args.output_directory)        


if __name__ == "__main__":
    main()
